/* detour.js - Framework for hooking TyranoScript
 * Copyright (C) 2016 Jaypee
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
(function (detour) {

  /* Adds a hook to a tag, to be called before the tag is started
   * and while attached.
   *
   * Hook callbacks take two arguments:
   *
   *   pm  - the tag parameters
   *   kag - the instance of kag that will start the tag
   *
   * There are no guarantees on the order in which hooks are run.
   */
  detour.addHookTagStart = function (tag, callback) {
    if (!this.kag.ftag.master_tag[tag]) {
      console.warn('addHookTagStart', 'no such tag');
    } else {
      this.tagStartHooks[tag] = this.tagStartHooks[tag] || [];
      this.tagStartHooks[tag].push(callback);
    }
  };

  /* Removes all hooks for the given tag, regardless of attachment.
   */
  detour.unhookTagStart = function (tag) {
    if (!this.kag.ftag.master_tag[tag]) {
      console.warn('unookTagStart', 'no such tag');
    } else {
      delete this.tagStartHooks[tag];
    }
  };

  /* Attaches a detour to the given tag, if not currently attached.
   *
   * The detour calls each of the hooks that have been added to
   * the tag before or after this call before starting the tag.
   */
  detour.attachTagStart = function (tag) {
    if (!this.kag.ftag.master_tag[tag]) {
      console.warn('attachTagStart', 'no such tag');
    } else if (this.tagStartBase[tag]) {
      console.warn('attachTagStart', 'attached already');
    } else {
      this.tagStartHooks[tag] = this.tagStartHooks[tag] || [];
      var master_tag = this.kag.ftag.master_tag[tag];
      this.tagStartBase[tag] = master_tag.start;
      master_tag.start = (function (base, pm) {
        this.tagStartHooks[tag].forEach(function (hook) {
          hook(pm, this.kag);
        }, this);
        base(pm);
      }).bind(this, master_tag.start.bind(master_tag));
    }
  };

  /* Detaches the detour from the given tag, if currently attached.
   *
   * Any hooks on the tag will not be called while detached.
   */
  detour.detachTagStart = function (tag) {
    if (!this.kag.ftag.master_tag[tag]) {
      console.warn('detachTagStart', 'no such tag');
    } else if (!this.tagStartBase[tag]) {
      console.warn('detachTagStart', 'detached already');
    } else {
      var master_tag = this.kag.ftag.master_tag[tag];
      master_tag.start = this.tagStartBase[tag];
      delete this.tagStartBase[tag];
    }
  };

  /* Adds a hook to parseScenario, to be called while attached.
   *
   * Hook callbacks take two arguments:
   *
   *   result_obj - the intermediate result from parseScenario
   *   kag        - the instance of kag that ran parseScenario
   *
   * If a hook modifies result_obj, then it must take care to keep
   * result_obj consistent internally. There are no guarantees on
   * the order in which hooks are run.
   */
  detour.addHookParseScenario = function (callback) {
    this.parseScenarioHooks.push(callback);
  };

  /* Removes all parseScenario hooks, regardless of attachment.
   */
  detour.unhookParseScenario = function () {
    this.parseScenarioHooks = [];
  };

  /* Attaches a detour to parseScenario, if not currently attached.
   *
   * The detour calls each of the hooks that have been added to
   * parseScenario before or after this call before resuming.
   */
  detour.attachParseScenario = function () {
    if (this.parseScenarioBase) {
      console.warn('attachParseScenario', 'attached already');
    } else {
      var parser = this.kag.parser;
      this.parseScenarioBase = parser.parseScenario;
      parser.parseScenario = (function (base, text_str) {
        var result_obj = base(text_str);
        this.parseScenarioHooks.forEach(function (hook) {
          hook(result_obj, this.kag);
          // Must update labels if tags are inserted or deleted,
          // but we don't expect clients to know about that.
          result_obj.array_s.forEach(updateLabelIndex);
        }, this);
        return result_obj;
      }).bind(this, parser.parseScenario.bind(parser));
    }
  };

  var updateLabelIndex = function (tag, i) {
    if (tag.name === 'label') {
      tag.pm.index = i;
    }
  };

  /* Detaches the detour from parseScenario, if currently attached.
   *
   * Any hooks on parseScenario will not be called while detached.
   */
  detour.detachParseScenario = function () {
    if (!this.parseScenarioBase) {
      console.warn('detachParseScenario', 'detached already');
    } else {
      this.kag.parser.parseScenario = this.parseScenarioBase;
      this.parseScenarioBase = null;
    }
  };

  /* Resets the framework, removing all current detours and hooks.
   */
  detour.reset = function () {
    this.init(this.kag);
  };

  /* Initializes the detour framework for the given instance of kag.
   *
   * Call this method first. It is safe to initialize more than one
   * instance of the framework with the same instance of kag only if
   * detours are always detached in reverse order of attachment.
   */
  detour.init = function (kag) {
    this.initTagStart();
    this.initParseScenario();
    this.kag = kag;
  };

  /* Initializes the framework state for tag start detours.
   */
  detour.initTagStart = function () {
    if (this.tagStartBase) {
      for (var tag in this.tagStartBase) {
        this.detachTagStart(tag);
      }
    } else {
      this.tagStartBase = {};
    }
    if (this.tagStartHooks) {
      for (var tag in this.tagStartHooks) {
        this.unhookTagStart(tag);
      }
    } else {
      this.tagStartHooks = {};
    }
  };

  /* Initializes the framework state for parseScenario detours.
   */
  detour.initParseScenario = function () {
    if (this.parseScenarioBase) {
      this.detachParseScenario();
    } else {
      this.parseScenarioBase = null;
    }
    if (this.parseScenarioHooks) {
      this.unhookParseScenario();
    } else {
      this.parseScenarioHooks = [];
    }
  };

})(window.TyranoDetour = window.TyranoDetour || {});
